
// memory usage with default params: ln((-0xffff) / ln(0.99) + 2) * (e / 0.01) = 4265.11707

/*****************************************************************************
	hCount for use with the new Tilted Time Windows with Subtraction
	01-NOV-2007
*****************************************************************************/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#include <assert.h>

#include <limits.h>
#define MAXINT INT_MAX

#include "histogram_hcsub.h"

extern long N;		// current transaction number
extern long BSIZE;	// batch size

/***************************************************************************** 
hc_type *hc_init(double rho, double epsilon, long M)
INPUTS:
	rho - hCount parameter
	epsilon - hCount parameter
	M - largest item in the stream

OUTPUT:
	Pointer to initialized hc_type struct

Creates an instance of the hc_type struct and initializes everything
*****************************************************************************/
histogram_hc_type *histogram_hc_init(double rho, double epsilon, long M)
{
    double e;			// min x/ln(x) and x > 1  --> evaluates to the base of the natural logarithm
    int R_RANGE;		// range for random numbers a_i and b_i
    int i;
    HISTOGRAM_TS_TYPE *aptr;

    histogram_hc_type *result = calloc(1, sizeof(histogram_hc_type));
    if (result == NULL) {
        printf("\ncalloc error");
        exit(100);
    }

    result->N = 0;			// total transactions
    result->M = M; 
    e = 2.718281828;
    //result->P = 1048583;	// big prime number
    //result->P = 1021;
    result->P = 65539;
    //result->P = 16411;

    R_RANGE = result->M;
    result->m =  e / epsilon;		// length of hash table 
    result->h = (int) (log(-result->M / log(rho)));		// width of hash table
    
    // Allocate memory for the counters S
    aptr = (HISTOGRAM_TS_TYPE *) malloc(result->m * result->h * sizeof(HISTOGRAM_TS_TYPE));
    if (aptr == NULL) {
        printf("\nmalloc() error.");
        exit(100);
    }
    result->S = (HISTOGRAM_TS_TYPE **) malloc(result->m *sizeof(HISTOGRAM_TS_TYPE));
    if (result->S == NULL) {
        printf("\nmalloc() error.");
        exit(200);
    }
    for (i = 0; i < result->m; i++) {
        result->S[i] = aptr + i * result->h;
    }
    
    // Set all counters to 0
    memset((void *) aptr, 0, result->m * result->h * sizeof(HISTOGRAM_TS_TYPE) / sizeof(char));
    
    // generate h pairs of random numbers a_i and b_i
    result->a = (int *) malloc(result->h * sizeof(int));
    if (result->a == NULL) {
        printf("\nmalloc() error.");
        exit(300);
    }
    result->b = (int *) malloc(result->h * sizeof(int));
    if (result->b == NULL) {
        printf("\nmalloc() error.");
        exit(400);
    }
    for (i = 0; i < result->h; i++) {
        result->a[i] = random(0, R_RANGE);
        result->b[i] = random(0, R_RANGE);
    }

    return (result);
}


//histogram_freq(i,4);

int histogram_value(double v, double min, double max, int nlevels) {


    // printf("histogram_value: mean=%f v=%f\n",mean,v);


    if(max == 0 && v == 0)
        return 0;

    //  printf("v = %f mean = %f\n", v, mean);
    assert(max != 0);
    assert(v <= max);

    //  printf("min=%f max=%f\n", min, max);

    if(nlevels == 2) {
        if(v < max* 0.5)
            return 0;
        return 1;
    }

    if(nlevels == 4) {
        if(v < max* 0.25)
            return 0;
        if(v < max * 0.5)
            return 1;
        if(v < max* 0.75)
            return 2;
        return 3;
    }

    if(nlevels == 8) {
        if(v < max* 0.125)
            return 0;
        if(v < max* 0.125*2)
            return 1;
        if(v < max* 0.125*3)
            return 2;
        if(v < max* 0.125*4)
            return 3;
        if(v < max* 0.125*5)
            return 4;
        if(v < max* 0.125*6)
            return 5;
        if(v < max* 0.125*7)
            return 6;
        return 7;
    }

    assert("wtf??" && 0);

    /*

    double normalized_max = max - min;
    double normalized_v = v - min;


    if(normalized_v == 0 && normalized_max == 0)
        return 0;

    assert(normalized_max > 0);

   // printf("norm v,max=%f,%f\n", normalized_v, normalized_max);



    if(normalized_v < normalized_max* 0.25) {
        return 0;
    }
    if(normalized_v < normalized_max * 0.5) {
        return 1;
    }
    if(normalized_v < normalized_max* 0.75) {
        return 2;
    }
    return 3;


*/

}


#define MAX(a,b) ((a) > (b) ? (a) : (b))
#define MIN(a,b) ((a) < (b) ? (a) : (b))

void histogram_update(histogram_hc_type * hc) {
    int j,i;
    int v;

    //  printf("performing histogram update for w0 and w1\n");
    // we update the histogram values of the 1st,2nd batches (we need both to update them)
    for (i=0; i<hc->m; i++) {
        for (j=0; j<hc->h; j++) {
            hc->S[i][j].histogram[0].f1=0;
            hc->S[i][j].histogram[0].f2=0;
            hc->S[i][j].histogram[0].f3=0;
            hc->S[i][j].histogram[0].f4=0;
            hc->S[i][j].histogram[0].f5=0;
            hc->S[i][j].histogram[0].f6=0;
            hc->S[i][j].histogram[0].f7=0;
            hc->S[i][j].histogram[0].f8=0;

            hc->S[i][j].histogram[1].f1=0;
            hc->S[i][j].histogram[1].f2=0;
            hc->S[i][j].histogram[1].f3=0;
            hc->S[i][j].histogram[1].f4=0;
            hc->S[i][j].histogram[1].f5=0;
            hc->S[i][j].histogram[1].f6=0;
            hc->S[i][j].histogram[1].f7=0;
            hc->S[i][j].histogram[1].f8=0;
        }
    }

}

/*****************************************************************************
void hc_update(hc_type * hc, int k, int ttype)
INPUTS:
	hc - pointer to the hCount struct
	k - the item in the stream
	ttype - transaction type: 1 for INSERT, -1 for DELETE

OUTPUT:
	NIL

Updates the hCount table with new transaction. 
*****************************************************************************/
void histogram_hc_update(histogram_hc_type * hc, int k, int ttype)
{
    int pos, j;

    hc->N++;
    N++;


    for (j = 0; j < hc->h; j++) {
        pos = ((((long long) hc->a[j] * k + hc->b[j]) % hc->P) % hc->m);
        if (ttype == INSERT) {
            hc->S[pos][j].c[0] = hc->S[pos][j].c[0] + 1;
        } else {
            hc->S[pos][j].c[0] = hc->S[pos][j].c[0] - 1;
        }
    }

    // Shift the Tilted Time Windows if necessary
    if (hc->N >= BSIZE && (hc->N-1)% BSIZE == 0) { // XXX -1 because N starts from 0.
        // printf("N = %d, performing shift.\n", hc->N);
        histogram_hc_shift(hc);
    }
}


void histogram4_shift(histogram_hc_type *hc, int powerValue) {
    int j,i,k;
    int pos;
    int hpos;


    assert(powerValue >= 1);

    //    printf("performing histogram shift from 0...%d to %d\n", powerValue, powerValue+1);

    // abbiamo 4*(powerValue+1) frequence per ciascun valore da accumulare in sole 4 frquence.
    // partendo dal passato, accumuliamo powerValue+1 (+1 perchè powervalue parte da 0)
    // frequence e quindi scriviamo una frequenza, poi resettiamo
    // l'accumulatore. cosi' ciascuna delle 4 freq. accumula powerValue frequenze.
    // le frequenze sono calcolate utilizzando sia l'istogramma che le freq nelle finestre precedenti.

    long n_items_seen;
    long n_items_count = 0;

    for(k = powerValue; k >= 0; k--) {
        n_items_count += (long)(BSIZE * pow(2.0, (k == 0 ? 1 : k)-1));
    }

    // printf("n_items_count = %d\n", n_items_count);

    for (i=0; i<hc->m; i++) {
        for (j=0; j<hc->h; j++) {

            // un item avrà sempre il solito hash, quindi qui non li considero esplicitamente.

            // ciclo sugli m valori possibili per ogni hashing function in h:

            n_items_seen = 0; // transazioni accumulate so far
            double c = 0; // accumulatore frequenze
            double c1 = 0;
            double c2 = 0;
            double c3 = 0;
            double c4 = 0;

            int set = 0; // serve solo per fare una assert

            pos = 4; // posizione frequenza da settare

            for(k = powerValue; k >= 0; k--) {

                // abbiamo powervalue+1 (0...powervalue estremi inclusi) finestre da 4 frequenze ciascuna (hpos)

                double hsum = hc->S[i][j].histogram[k].f1+
                              hc->S[i][j].histogram[k].f2+
                              hc->S[i][j].histogram[k].f3+
                              hc->S[i][j].histogram[k].f4;

                hpos = 4;

                for(;;) {

                    assert("prima" && c >= 0);

                    //  printf("c = %d * %d / %f\n",hc->S[i][j].c[k], hc->S[i][j].histogram[k].f1, hsum);

                    if(hsum) {
                        switch(hpos) {
                        case 1:
                            c += ((double)(hc->S[i][j].c[k])) * (((double)(hc->S[i][j].histogram[k].f1)) / hsum);
                            break;
                        case 2:
                            c += ((double)(hc->S[i][j].c[k])) * (((double)(hc->S[i][j].histogram[k].f2)) / hsum);
                            break;
                        case 3:
                            c += ((double)(hc->S[i][j].c[k])) * (((double)(hc->S[i][j].histogram[k].f3)) / hsum);
                            break;
                        case 4:
                            c += ((double)(hc->S[i][j].c[k])) * (((double)(hc->S[i][j].histogram[k].f4)) / hsum);
                            break;
                        default:
                            assert(0);
                        }
                    } else { // if everything is at level 0 in histogram, assume uniform distrib.
                        c+= ((double)(hc->S[i][j].c[k])) * 0.25;
                    }
                    hpos--;

                    assert("dopo" && c >= 0);


                    assert((long)(BSIZE * pow(2.0, (k == 0 ? 1 : k)-1)) % 4 == 0);
                    assert(n_items_count % 4 == 0);

                    n_items_seen += (long)(BSIZE * pow(2.0, (k == 0 ? 1 : k)-1))/4;

                    assert(pos > 0);

                    if(n_items_seen >= n_items_count/4) {
                        //  printf("CASE: n_items_seen >= n_items_count/4\n");
                        //  if(i == 0 && j == 0)
                        //  printf("setting h %d\n", pos);

                        //     int mean = hc->S[i][j].c[powerValue + 1]/4;

                        switch(pos) {
                        case 1:
                            set++;
                            c1 = c;
                            long max = MAX(c1, MAX(c2, MAX(c3, c4))); // cera un +1 ma credo fosse un errore!
                            long min = MIN(c1, MIN(c2, MIN(c3, c4)));

                            //       printf("c1=%f c2=%f c3=%f c4=%f max=%ld aggregatedsum=%ld\n", c1,c2,c3,c4,max, hc->S[i][j].c[powerValue+1]);
                            assert(abs((c1+c2+c3+c4) - hc->S[i][j].c[powerValue+1]) < 1);
                            // approximation error < 1!
                            hc->S[i][j].histogram[powerValue + 1].f1 = histogram_value((int)c1, min,max,4);
                            hc->S[i][j].histogram[powerValue + 1].f2 = histogram_value((int)c2, min,max,4);
                            hc->S[i][j].histogram[powerValue + 1].f3 = histogram_value((int)c3, min,max,4);
                            hc->S[i][j].histogram[powerValue + 1].f4 = histogram_value((int)c4, min,max,4);

                            break;
                        case 2:
                            c2 = c;
                            break;
                        case 3:
                            c3 = c;
                            break;
                        case 4:
                            c4 = c;
                            break;
                        default:
                            printf("err:  pos=%d k=%d\n",pos,k);
                            assert(0);
                        }
                        pos--;
                        n_items_seen = 0;
                        c = 0;
                    }

                    if(hpos < 1) // start accumulating in next (more recent) window
                        break;
                }
            }
            assert(hpos == 0 && pos == 0 && set == 1);
        }
    }
}



void histogram2_shift(histogram_hc_type *hc, int powerValue) {
    int j,i,k;
    int pos;
    int hpos;


    assert(powerValue >= 1);

    //    printf("performing histogram shift from 0...%d to %d\n", powerValue, powerValue+1);

    // abbiamo 4*(powerValue+1) frequence per ciascun valore da accumulare in sole 4 frquence.
    // partendo dal passato, accumuliamo powerValue+1 (+1 perchè powervalue parte da 0)
    // frequence e quindi scriviamo una frequenza, poi resettiamo
    // l'accumulatore. cosi' ciascuna delle 4 freq. accumula powerValue frequenze.
    // le frequenze sono calcolate utilizzando sia l'istogramma che le freq nelle finestre precedenti.

    long n_items_seen;
    long n_items_count = 0;

    for(k = powerValue; k >= 0; k--) {
        n_items_count += (long)(BSIZE * pow(2.0, (k == 0 ? 1 : k)-1));
    }

    // printf("n_items_count = %d\n", n_items_count);

    for (i=0; i<hc->m; i++) {
        for (j=0; j<hc->h; j++) {

            // un item avrà sempre il solito hash, quindi qui non li considero esplicitamente.

            // ciclo sugli m valori possibili per ogni hashing function in h:

            n_items_seen = 0; // transazioni accumulate so far
            double c = 0; // accumulatore frequenze
            double c1 = 0;
            double c2 = 0;


            int set = 0; // serve solo per fare una assert

            pos = 2; // posizione frequenza da settare

            for(k = powerValue; k >= 0; k--) {

                // abbiamo powervalue+1 (0...powervalue estremi inclusi) finestre da 4 frequenze ciascuna (hpos)

                double hsum = hc->S[i][j].histogram[k].f1+
                              hc->S[i][j].histogram[k].f2;

                hpos = 2;

                for(;;) {

                    assert("prima" && c >= 0);

                    //  printf("c = %d * %d / %f\n",hc->S[i][j].c[k], hc->S[i][j].histogram[k].f1, hsum);

                    if(hsum) {
                        switch(hpos) {
                        case 1:
                            c += ((double)(hc->S[i][j].c[k])) * (((double)(hc->S[i][j].histogram[k].f1)) / hsum);
                            break;
                        case 2:
                            c += ((double)(hc->S[i][j].c[k])) * (((double)(hc->S[i][j].histogram[k].f2)) / hsum);
                            break;
                        default:
                            assert(0);
                        }
                    } else { // if everything is at level 0 in histogram, assume uniform distrib.
                        c+= ((double)(hc->S[i][j].c[k])) * 0.5;
                    }
                    hpos--;

                    assert("dopo" && c >= 0);


                    assert((long)(BSIZE * pow(2.0, (k == 0 ? 1 : k)-1)) % 2 == 0);
                    assert(n_items_count % 2 == 0);

                    n_items_seen += (long)(BSIZE * pow(2.0, (k == 0 ? 1 : k)-1))/2;

                    assert(pos > 0);

                    if(n_items_seen >= n_items_count/2) {
                        //  printf("CASE: n_items_seen >= n_items_count/4\n");
                        //  if(i == 0 && j == 0)
                        //  printf("setting h %d\n", pos);

                        //     int mean = hc->S[i][j].c[powerValue + 1]/4;

                        switch(pos) {
                        case 1:
                            set++;
                            c1 = c;
                            long max = MAX(c1, c2);
                            long min = MIN(c1, c2);

                            //       printf("c1=%f c2=%f c3=%f c4=%f max=%ld aggregatedsum=%ld\n", c1,c2,c3,c4,max, hc->S[i][j].c[powerValue+1]);
                            assert(abs((c1+c2) - hc->S[i][j].c[powerValue+1]) < 1);
                            // approximation error < 1!
                            hc->S[i][j].histogram[powerValue + 1].f1 = histogram_value((int)c1, min,max,8);
                            hc->S[i][j].histogram[powerValue + 1].f2 = histogram_value((int)c2, min,max,8);

                            break;
                        case 2:
                            c2 = c;
                            break;
                        default:
                            printf("err:  pos=%d k=%d\n",pos,k);
                            assert(0);
                        }
                        pos--;
                        n_items_seen = 0;
                        c = 0;
                    }

                    if(hpos < 1) // start accumulating in next (more recent) window
                        break;
                }
            }
            assert(hpos == 0 && pos == 0 && set == 1);
        }
    }
}



void histogram8_shift(histogram_hc_type *hc, int powerValue) {
    int j,i,k;
    int pos;
    int hpos;


    assert(powerValue >= 1);

    //    printf("performing histogram shift from 0...%d to %d\n", powerValue, powerValue+1);

    // abbiamo 4*(powerValue+1) frequence per ciascun valore da accumulare in sole 4 frquence.
    // partendo dal passato, accumuliamo powerValue+1 (+1 perchè powervalue parte da 0)
    // frequence e quindi scriviamo una frequenza, poi resettiamo
    // l'accumulatore. cosi' ciascuna delle 4 freq. accumula powerValue frequenze.
    // le frequenze sono calcolate utilizzando sia l'istogramma che le freq nelle finestre precedenti.

    long n_items_seen;
    long n_items_count = 0;

    for(k = powerValue; k >= 0; k--) {
        n_items_count += (long)(BSIZE * pow(2.0, (k == 0 ? 1 : k)-1));
    }

    // printf("n_items_count = %d\n", n_items_count);

    for (i=0; i<hc->m; i++) {
        for (j=0; j<hc->h; j++) {

            // un item avrà sempre il solito hash, quindi qui non li considero esplicitamente.

            // ciclo sugli m valori possibili per ogni hashing function in h:

            n_items_seen = 0; // transazioni accumulate so far
            double c = 0; // accumulatore frequenze
            double c1 = 0;
            double c2 = 0;
            double c3 = 0;
            double c4 = 0;
            double c5 = 0;
            double c6 = 0;
            double c7 = 0;
            double c8 = 0;

            int set = 0; // serve solo per fare una assert

            pos = 8; // posizione frequenza da settare

            for(k = powerValue; k >= 0; k--) {

                // abbiamo powervalue+1 (0...powervalue estremi inclusi) finestre da 4 frequenze ciascuna (hpos)

                double hsum = hc->S[i][j].histogram[k].f1+
                              hc->S[i][j].histogram[k].f2+
                              hc->S[i][j].histogram[k].f3+
                              hc->S[i][j].histogram[k].f4+
                              hc->S[i][j].histogram[k].f5+
                              hc->S[i][j].histogram[k].f6+
                              hc->S[i][j].histogram[k].f7+
                              hc->S[i][j].histogram[k].f8;

                hpos = 8;

                for(;;) {

                    assert("prima" && c >= 0);

                    //  printf("c = %d * %d / %f\n",hc->S[i][j].c[k], hc->S[i][j].histogram[k].f1, hsum);

                    if(hsum) {
                        switch(hpos) {
                        case 1:
                            c += ((double)(hc->S[i][j].c[k])) * (((double)(hc->S[i][j].histogram[k].f1)) / hsum);
                            break;
                        case 2:
                            c += ((double)(hc->S[i][j].c[k])) * (((double)(hc->S[i][j].histogram[k].f2)) / hsum);
                            break;
                        case 3:
                            c += ((double)(hc->S[i][j].c[k])) * (((double)(hc->S[i][j].histogram[k].f3)) / hsum);
                            break;
                        case 4:
                            c += ((double)(hc->S[i][j].c[k])) * (((double)(hc->S[i][j].histogram[k].f4)) / hsum);
                            break;
                        case 5:
                            c += ((double)(hc->S[i][j].c[k])) * (((double)(hc->S[i][j].histogram[k].f5)) / hsum);
                            break;
                        case 6:
                            c += ((double)(hc->S[i][j].c[k])) * (((double)(hc->S[i][j].histogram[k].f6)) / hsum);
                            break;
                        case 7:
                            c += ((double)(hc->S[i][j].c[k])) * (((double)(hc->S[i][j].histogram[k].f7)) / hsum);
                            break;
                        case 8:
                            c += ((double)(hc->S[i][j].c[k])) * (((double)(hc->S[i][j].histogram[k].f8)) / hsum);
                            break;
                        default:
                            assert(0);
                        }
                    } else { // if everything is at level 0 in histogram, assume uniform distrib.
                        c+= ((double)(hc->S[i][j].c[k])) * 0.125;
                    }
                    hpos--;

                    assert("dopo" && c >= 0);


                    assert((long)(BSIZE * pow(2.0, (k == 0 ? 1 : k)-1)) % 8 == 0);
                    assert(n_items_count % 8 == 0);

                    n_items_seen += (long)(BSIZE * pow(2.0, (k == 0 ? 1 : k)-1))/8;

                    assert(pos > 0);

                    if(n_items_seen >= n_items_count/8) {
                        //  printf("CASE: n_items_seen >= n_items_count/4\n");
                        //  if(i == 0 && j == 0)
                        //  printf("setting h %d\n", pos);

                        //     int mean = hc->S[i][j].c[powerValue + 1]/4;

                        switch(pos) {
                        case 1:
                            set++;
                            c1 = c;
                            long max = MAX(c1, MAX(c2, MAX(c3,MAX(c4, MAX(c5, MAX(c6, MAX(c7, c8))))))); // cera un +1 ma credo fosse un errore!
                            long min = MIN(c1, MIN(c2, MIN(c3,MIN(c4, MIN(c5, MIN(c6, MIN(c7, c8))))))); // cera un +1 ma credo fosse un errore!


                            //printf("c1=%f c2=%f c3=%f c4=%f max=%ld aggregatedsum=%ld\n", c1,c2,c3,c4,max, hc->S[i][j].c[powerValue+1]);
                            assert(abs((c1+c2+c3+c4+c5+c6+c7+c8) - hc->S[i][j].c[powerValue+1]) < 1);
                            // approximation error < 1!
                            hc->S[i][j].histogram[powerValue + 1].f1 = histogram_value((int)c1, min,max,2);
                            hc->S[i][j].histogram[powerValue + 1].f2 = histogram_value((int)c2, min,max,2);
                            hc->S[i][j].histogram[powerValue + 1].f3 = histogram_value((int)c3, min,max,2);
                            hc->S[i][j].histogram[powerValue + 1].f4 = histogram_value((int)c4, min,max,2);

                            hc->S[i][j].histogram[powerValue + 1].f5 = histogram_value((int)c5, min,max,2);
                            hc->S[i][j].histogram[powerValue + 1].f6 = histogram_value((int)c6, min,max,2);
                            hc->S[i][j].histogram[powerValue + 1].f7 = histogram_value((int)c7, min,max,2);
                            hc->S[i][j].histogram[powerValue + 1].f8 = histogram_value((int)c8, min,max,2);
                            break;
                        case 2:
                            c2 = c;
                            break;
                        case 3:
                            c3 = c;
                            break;
                        case 4:
                            c4 = c;
                            break;
                        case 5:
                            c5 = c;
                            break;
                        case 6:
                            c6 = c;
                            break;
                        case 7:
                            c7 = c;
                            break;
                        case 8:
                            c8 = c;
                            break;

                        default:
                            printf("err:  pos=%d k=%d\n",pos,k);
                            assert(0);
                        }
                        pos--;
                        n_items_seen = 0;
                        c = 0;
                    }

                    if(hpos < 1) // start accumulating in next (more recent) window
                        break;
                }
            }
            assert(hpos == 0 && pos == 0 && set == 1);
        }
    }
}
/*****************************************************************************
void hc_shift(hc_type *hc)
INPUTS:
	hc - pointer to the hCount struct
	
OUTPUT:
	NIL

Shifts the Tilted Time Windows associated with each counter
*****************************************************************************/



void histogram_hc_shift(histogram_hc_type *hc)
{
    int i, j, k;
    int index, powerValue, threshold;


    index = hc->N/BSIZE;

    int onlyonce = 1;

    for (powerValue = WINDOWS-2; powerValue >= 0; powerValue--) {
        threshold = (int)pow(2.0, (double)powerValue);
        if (hc->N > 0 && index % threshold == 0) {

            // printf("shift a N=%d powerValue=%d\n", hc->N, powerValue);

            for (i=0; i<hc->m; i++) {
                for (j=0; j<hc->h; j++) {
                    hc->S[i][j].c[powerValue + 1] = 0;
                    for(k = 0; k < powerValue + 1; k++) {
                        hc->S[i][j].c[powerValue + 1] += hc->S[i][j].c[k];
                    }
                }
            }

            if(powerValue > 0) { // if not just a shift from w0 to w1...
                if(onlyonce) {
                    histogram_update(hc);
                    onlyonce = 0;
                }
                histogram4_shift(hc, powerValue);
            }


            //   printf("cumulating windows [0..%d] => %d\n", powerValue,powerValue+1);

            //   printf("shiftbegin %d\n",powerValue);




        }
    }

    /*
    for (powerValue = WINDOWS-2; powerValue >= 0; powerValue--) {
        threshold = (int)pow(2.0, (double)powerValue);
        if (hc->N > 0 && index % threshold == 0) {
            histogram_shift(hc, powerValue);
        }
    }
*/

    for (i=0; i<hc->m; i++) {
        for (j=0; j<hc->h; j++) {
            hc->S[i][j].c[0] = 0;
        }
    }

}



/*****************************************************************************
long hc_query(hc_type *hc, int id, long item)	
INPUTS:
	hc - pointer to the hCount struct
	id - ID of the Tilted Time Window
	item - item to look for

OUTPUT:
	Count of "item" in window number "id"

Counts the number of occurences of "item" in window number "id" (This is the
core query part of the hCount algorithm)
*****************************************************************************/

long histogram_hc_query(histogram_hc_type *hc, int id, long item)
{
    long min = MAXINT;
    int j, pos, c;

    for (j = 0; j < hc->h; j++) {
        pos = ((((long long) hc->a[j] * item + hc->b[j]) % hc->P) % hc->m);
        c = hc->S[pos][j].c[id];
        if (c < min) {
            min = c;
        }
    }

    return(min);
}


/*
long histogram_hc_query(histogram_hc_type *hc, int id, long item)
{
    int j, pos, c;
    histogram_struct *s = NULL;

    for (j = 0; j < hc->h; j++) {
        pos = ((((long long) hc->a[j] * item + hc->b[j]) % hc->P) % hc->m);

        if(!s) {
            s = &hc->S[pos][j].histogram[id];
            c = hc->S[pos][j].c[id];
            continue;
        }

        if(hc->S[pos][j].histogram[id].f1 < s->f1 &&
           hc->S[pos][j].histogram[id].f2 < s->f2 &&
           hc->S[pos][j].histogram[id].f3 < s->f3 &&
           hc->S[pos][j].histogram[id].f4 < s->f4) {
            s = &hc->S[pos][j].histogram[id];
            c = hc->S[pos][j].c[id];
        }


    }

    return(c);
}*/

/*****************************************************************************
void hc_destroy(hc_type * hc)
INPUTS:
	hc - pointer to the hCount structure

OUTPUT:
	NIL

Frees up memory used by the hCount structure
*****************************************************************************/
void histogram_hc_destroy(histogram_hc_type * hc)
{
    free(hc->a);
    free(hc->b);
    free(hc->S);
}

