
/*****************************************************************************
*****************************************************************************/



#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <getopt.h>
// headers from massdal code
#include "prng.h"
#include "massdal.h"

// new headers for titicount code
#include "hcsub.h"
#include "sub.h"
#include "hc2.h"
#include "genquery.h"
#include "linear.h"
#include "linearb.h"
#include "ssw.h"
#include "ssw2.h"
#include "histogram_hcsub.h"
#include "histogram_sub.h"

#include <math.h>               // Needed for pow()

//----- Constants -----------------------------------------------------------
#define  FALSE          0       // Boolean false
#define  TRUE           1       // Boolean true

//----- Function prototypes -------------------------------------------------
int      zipf(double alpha, int n);  // Returns a Zipf random variable
double   rand_val(int seed);         // Jain's RNG



void  initOpt(int argc, char **argv) ;
int mainloop();
int mainloop2();
void get_cmd_params(int argc, char **argv);
unsigned int *alg2_output(hc2_type *hc2, long tqmin, long tqmax);
unsigned int *alg3_output(hc_type *titi, long tqmin, long tqmax);
unsigned int *alg4_output(linear_type *lc, long tqmin, long tqmax);
unsigned int *alg5_output(linearb_type *lc, long tqmin, long tqmax);
unsigned int *alg6_output(SSW_type *ss, long tqmin, long tqmax);
unsigned int *alg7_output(SSW2_type *ss2, long tqmin, long tqmax);
unsigned int *alg8_output(histogram_hc_type *titi, long tqmin, long tqmax);

void CheckOutput(const char *title, unsigned int *resultlist, int hh, int upt, int outt);
int *ReadStream();
int *ReadStream2(char *fname, int length);

void experiment_topk( long tqmin, long tqmax) ;
void experiment_freqs( long tqmin, long tqmax) ;
void experiment_topk_freqs();
void experiment_zipf();
void experiment_time();
int *alg6_topk(SSW_type *ssw, int k,long tqmin, long tqmax) ;
void sort(unsigned int *counters) ;



long N = 0;
long LBBSIZE = 100000; // used by linearb.c
int beta = 0;
int *stream;
int *exact;
long M = 0xffff;
long rseed = 0;


long BSIZE = -1; // redundant, there is already o.BSIZE. however, left because used in other files.
// setted in initOpt, always equal to o.BSIZE.


/*long range = 1000000;
long BSIZE = 10000;
double thresh = 0.005;
long tqmin, tqmax;
//int QWIDTH;
int CHANGING_DATA = TRUE;
int RUNS = 5;
double zipfpar = 1.1;
char *fdata;*/


typedef struct {
    long range;
    long BSIZE;
    double thresh;
    long tqmin;
    long tqmax;
    int QWIDTH;
    int CHANGING_DATA;
    int RUNS;
    double zipfpar;
    char *fdata;
    char *sdata;
    int experimentId;
} OPT;


OPT o;





/******************************************************************************
Calculate exact frequencies

******************************************************************************/
long ExactWin(long tqmin, long tqmax)
{
    long i, cnt, val;
    long hh=0;
    long cutoff = o.thresh * (tqmax - tqmin);

    //printf("\nA_E_List:");
    for (val=1; val<M; val++) {
        cnt = 0;
        for (i=tqmin; i<=tqmax; i++) {
            if (stream[i] == val) {
           //     printf("stream offset=%d id=%d\n", i, stream[i]);
                cnt++;
            }
        }
        exact[val] = cnt;
        if (cnt > cutoff) {
            hh++;
            //printf("\nA_E: %ld:%ld", val, cnt);
        }
    }

//    for (val=1; val<M; val++) {
 //       printf("ground-truth id=%d freq=%d\n", val, exact[val]);
  //  }

    return(hh);
}


int *CreateStream3(int length) {
    char str[512];
    long long i,j;
    int *stream;
    FILE *fin;

    // srand((unsigned) time(NULL));
    srand(0);

    sprintf(str, "%s/zipf-%d-%.1f", o.sdata, rand() % 20+1, o.zipfpar);

    printf("reading zipf distribution from %s ...\n", str);

    fin=fopen(str, "r");
    assert(fin != NULL);

    exact = (int *) calloc(M + 1, sizeof(int));
    stream = (int *) calloc(o.range + 1, sizeof(int));
    if (exact==NULL || stream==NULL) {
        printf("calloc error");
        exit(123);
    }



    long a, b;
    long value;
    long long lim1, lim2;
    prng_type *prng;
    long long netpos;

    prng = prng_Init(0, 2); //prng_Init(time(NULL), 2);

    netpos = 0;

    beta = 5;// zipfpar;

    long div = length / (o.BSIZE*beta);
    printf("using createstream3 with beta=%d. switching %d times\n", beta, div);

    if(div < 1)
        div = 1;

    if (o.CHANGING_DATA) {
        for (j=0; j<div; j++) {
            a = (long long) (prng_int(prng) % MOD);
            b = (long long) (prng_int(prng) % MOD);

            lim1 = j*length/div + 1;
            lim2 = lim1 + length/div;

            for (i = lim1; i <= lim2; i++) {

                fscanf(fin, "%s", str);
                value = atoi(str);
                value = (hash31(a, b, atoi(str)) & M);

                exact[value]++;
                netpos++;

                if(i > length) {
                    printf("trying to write after end: %d, skipping\n", i); // XXX this is due to a bug of the data stream generator and happens exactly after the last "correct" position. it is a "offset+1 bug" problem
                    continue;
                }

                stream[i] = value;
            }
        }

    } else {

        a = (long long) (prng_int(prng) % MOD);
        b = (long long) (prng_int(prng) % MOD);

        for (i = 1; i <= length; i++) {

            fscanf(fin, "%s", str);
            value = atoi(str);
            value = (hash31(a, b, atoi(str)) & M);

            exact[value]++;
            netpos++;

            stream[i] = value;
        }
    }

    prng_Destroy(prng);
    return (stream);

}

/******************************************************************************
Generate synthetic data set

******************************************************************************/
int *CreateStream(int length)
{
    long a, b;
    double zet;
    long i, j;
    long value;
    long lim1, lim2;
    int *stream;
    prng_type *prng;
    int netpos;

    exact = (int *) calloc(M + 1, sizeof(int));
    stream = (int *) calloc(length + 1, sizeof(int));

    assert(rseed == 0);

    if (rseed == 0) {
        prng = prng_Init(time(NULL), 2);
        srand((unsigned) time(NULL));
    } else {
        prng = prng_Init(rseed, 2);
        srand(rseed);
    }

    netpos = 0;


    int div = length / (5*o.BSIZE);

    if(div < 1)
        div = 1;

    printf("div=%d\n", div);

    if (o.CHANGING_DATA) {
        zet = zeta(length/div, o.zipfpar);
        for (j=0; j<div; j++) {
            a = (long long) (prng_int(prng) % MOD);
            b = (long long) (prng_int(prng) % MOD);

            lim1 = j*length/div + 1;
            lim2 = lim1 + length/div;

            for (i = lim1; i <= lim2; i++) {
                value = (hash31(a, b, ((int) floor(fastzipf(o.zipfpar, M, zet, prng)))) & M);

                exact[value]++;
                netpos++;

                if(i > length) {
                    printf("trying to write after end: %d, skipping\n", i); // XXX this is due to a bug of the data stream generator and happens exactly after the last "correct" position. it is a "offset+1 bug" problem
                    continue;
                }

                stream[i] = value;
            }
        }

    } else {
        zet = zeta(length, o.zipfpar);

        a = (long long) (prng_int(prng) % MOD);
        b = (long long) (prng_int(prng) % MOD);

        for (i = 1; i <= length; i++) {
            value = (hash31(a, b, ((int) floor(fastzipf(o.zipfpar, M, zet, prng)))) & M);
            exact[value]++;
            netpos++;

            stream[i] = value;
        }
    }

    prng_Destroy(prng);
    return (stream);

}

// version of Michele Dallachiesa
int *CreateStream2(int length)
{
    long a, b;
    double zet;
    long i, j;
    long value;
    long lim1, lim2;
    int *stream;
    prng_type *prng;
    int netpos;

    exact = (int *) calloc(M + 1, sizeof(int));
    stream = (int *) calloc(length + 1, sizeof(int));

    netpos = 0;

    prng = prng_Init(time(NULL), 2);
    srand((unsigned) time(NULL));



    int div = length / (5*o.BSIZE);

    if(div < 1)
        div = 1;

    printf("div=%d\n", div);

    rand_val(time(NULL));
    zipf(0,0);


    if (o.CHANGING_DATA) {
        for (j=0; j<div; j++) {
            a = (long long) (prng_int(prng) % MOD);
            b = (long long) (prng_int(prng) % MOD);

            lim1 = j*length/div + 1;
            lim2 = lim1 + length/div;

            for (i = lim1; i <= lim2; i++) {

                //value = zipf(zipfpar, M);
                value = (hash31(a, b, ((int) floor(zipf(o.zipfpar, M)))) & M);

                exact[value]++;
                netpos++;

                if(i > length) {
                    printf("trying to write after end: %d, skipping\n", i); // XXX this is due to a bug of the data stream generator and happens exactly after the last "correct" position. it is a "offset+1 bug" problem
                    continue;
                }

                stream[i] = value;
            }
        }

    } else {
        a = (long long) (prng_int(prng) % MOD);
        b = (long long) (prng_int(prng) % MOD);

        for (i = 1; i <= length; i++) {
            value = (hash31(a, b, ((int) floor(zipf(o.zipfpar, M)))) & M);
            //  value = zipf(zipfpar, M);
            //  printf("value = %d\n", value);
            exact[value]++;
            netpos++;
            stream[i] = value;
        }
    }


    prng_Destroy(prng);
    return (stream);
}


int main(int argc, char *argv[])
{
    int i;


    srand(0);
    // srand((unsigned)time(NULL));

    initOpt(argc, argv);

    //  get_cmd_params(argc, argv);


    for (i=0; i<o.RUNS; i++) {
        mainloop();			// for queries specified using qmin, qmax in input file
        //  mainloop2();		// for random queries
    }

    return(0);
}

/******************************************************************************
	mainloop2()
	Random queries

	Tests algorithms with random queries of width QWIDTH
******************************************************************************/
int mainloop2()
{
    long hh;
    long uptime3, outtime2, outtime3, outtime4, outtime5;
    double dummy;
    unsigned int *uilist;
    int i, j;
    long rangeby10 = o.range/3;
    //int QWIDTH = CHANGING_DATA;		// CHANGING_DATA is used as width in this expt.

    if (o.fdata) {
        stream = ReadStream2(o.fdata,o.range);
    } else {
        stream = CreateStream(o.range);
    }

    N = 0;
    hc2_type *hc;
    hc_type *titi;
    linear_type *lc;
    linearb_type *lcb;

    hc = hc2_init(0.99, 0.01, M);
    titi = hc_init(0.99, 0.01, M);
    lc = linear_init(0.99, 0.01, M);
    lcb = linearb_init(0.99, 0.01, M);

    StartTheClock();
    for (i=1; i<o.range; i++) {

        linearb_update(lcb, stream[i], INSERT);
        N--;
        hc2_update(hc, stream[i], INSERT);
        N--;
        hc_update(titi, stream[i], INSERT);
        N--;
        linear_update(lc, stream[i], INSERT);

        if (i % rangeby10 == 0) {
            uptime3 = StopTheClock();

            for (j=0; j<10; j++) {
                gen_query(i, &o.tqmin, &o.tqmax, &dummy, o.QWIDTH);
                //thresh = 0.01;
                printf("\nN=%d: (%ld, %ld) (%f)", i, o.tqmin, o.tqmax, o.thresh);

                hh = ExactWin(o.tqmin, o.tqmax);

                StartTheClock();
                uilist = alg2_output(hc, o.tqmin, o.tqmax);
                outtime2 = StopTheClock();
                CheckOutput("TiTi", uilist, hh, uptime3, outtime2);
                free(uilist);

                StartTheClock();
                uilist = alg3_output(titi, o.tqmin, o.tqmax);
                outtime3 = StopTheClock();
                CheckOutput("TiTi+", uilist, hh, uptime3, outtime3);
                free(uilist);


                StartTheClock();
                uilist = alg4_output(lc, o.tqmin, o.tqmax);
                outtime4 = StopTheClock();
                CheckOutput("Lin", uilist, hh, uptime3, outtime4);
                free(uilist);

                //gen_query(i, &tqmin, &tqmax, &dummy, QWIDTH*100);
                StartTheClock();
                uilist = alg5_output(lcb, o.tqmin, o.tqmax);
                outtime5 = StopTheClock();
                CheckOutput("BLin", uilist, hh, uptime3, outtime5);
                free(uilist);

            }

            StartTheClock();
        }
    }
    

    hc_destroy(titi);
    hc2_destroy(hc);
    linear_destroy(lc);
    linearb_destroy(lcb);

    free(stream);
    free(exact);

    return(0);
}

long get_history_size() {
    int i;
    long history_size = o.BSIZE;

    for(i = 0; i <WINDOWS-1; i++) {
        history_size += pow(2,i-1) * o.BSIZE;
    }
    return history_size;
}

/******************************************************************************
	mainloop()
	User defined queries

	Tests algorithms with queries defined by tqmin, tqmax
******************************************************************************/
int mainloop()
{


    printf("amnesic model size: %ld (bsize = %d, windows = %d)\n", get_history_size(), o.BSIZE, WINDOWS);

    if (o.fdata) {
        stream = ReadStream(o.fdata,o.range);
    } else {
        assert(o.sdata);
        stream = CreateStream3(o.range);
    }


    switch(o.experimentId) {
    case 11:
        experiment_topk_freqs();
        break;
    case 9:
        experiment_freqs(o.tqmin, o.tqmax);
    case 101:
        experiment_freqs(o.tqmin, o.tqmax);
        break;
    default:
        assert("WTF!?" && 0);
    }

    // experiment_topk(tqmin,tqmax);
    //  experiment_freqs(tqmin, tqmax);
    //     experiment_zipf();
    // experiment_time();

    free(stream);
    free(exact);

    return(0);
}


void  initOpt(int argc, char **argv) {
    int            c;

    printf("*** cmd: ");
    int i;
    for(i = 0; i < argc; i++) {
        printf("%s ", argv[i]);
    }
    printf("***\n");



    o.range = 100000;
    o.BSIZE = 1000;
    o.thresh = 0.005;
    o.tqmin = 1;
    o.tqmax = o.range;
    o.CHANGING_DATA = 1;
    o.RUNS = 1;
    o.zipfpar = 1.1;
    o.fdata = NULL;
    o.sdata = NULL;
    o.experimentId = -1;

    if (argc ==1){
        printf("err\n");
        exit(1);
    }

    opterr = 0;

    while ((c = getopt(argc, argv, "r:B:t:m:M:C:R:Z:f:s:e:")) != EOF)
        switch (c)
        {

        case 'r':
        o.range = atoi(optarg);
        break;

        case 'B':
        o.BSIZE = atoi(optarg);
        break;

        case 't':
        o.thresh = atof(optarg);
        break;

        case 'm':
        o.tqmin = atoi(optarg);
        break;

        case 'M':
        o.tqmax = atoi(optarg);
        break;

        case 'C':
        o.CHANGING_DATA = atoi(optarg); // 0 = false, 1 = true
        break;

        case 'R':
        o.RUNS = atoi(optarg);
        break;

        case 'Z':
        o.zipfpar = atof(optarg);
        break;

        case 'f':
        o.fdata = strdup(optarg);
        break;

        case 's':
        o.sdata = strdup(optarg);
        break;

        case 'e':
        o.experimentId = atoi(optarg);
        break;

        default:
        printf("err: option '%c' invalid\n", optopt);
        exit(1);
    }

    o.QWIDTH = o.tqmin;

    BSIZE = o.BSIZE;


    printf("\n#N %ld \
           \nBSIZE %ld \
           \nThreshold %f \
           \ntqmin %ld \
           \ntqmax %ld \
           \nCHANGING_DATA %d \
           \nRUNS %d \
           \nzipf %f\n",
           o.range, o.BSIZE, o.thresh, o.tqmin, o.tqmax, o.CHANGING_DATA, o.RUNS, o.zipfpar);

}



/******************************************************************************
Get parameters from command line. 
Very basic. Order of parameters is important.

******************************************************************************/
/*void get_cmd_params(int argc, char **argv)
{





    if (argc > 1)
        range = atoi(argv[1]);
    else
        range = 100000;

    if (argc > 2)
        BSIZE = atoi(argv[2]);
    else
        BSIZE = 1000;

    if (argc > 3)
        thresh = atof(argv[3]);
    else
        thresh = 0.005;

    if (argc > 4)
        tqmin = atoi(argv[4]);
    else
        tqmin = 1;

    QWIDTH = tqmin;

    if (argc > 5)
        tqmax = atoi(argv[5]);
    else
        tqmax = range;

    if (argc > 6)
        CHANGING_DATA = atoi(argv[6]);
    else
        CHANGING_DATA = 1;

    if (argc > 7)
        RUNS = atoi(argv[7]);
    else
        RUNS = 1;

    if (argc > 8) {
        zipfpar = atof(argv[8]);
    } else
        zipfpar = 1.1;

    if (argc > 9) {
        fdata = argv[9];
        printf("#got file %s\n", fdata);
    } else
        fdata = NULL;



    printf("\n#N %ld \
           \nBSIZE %ld \
           \nThreshold %f \
           \ntqmin %ld \
           \ntqmax %ld \
           \nCHANGING_DATA %d \
           \nRUNS %d \
           \nzipf %f\n",
           range, BSIZE, thresh, tqmin, tqmax, CHANGING_DATA, RUNS, zipfpar);
}*/

unsigned int *alg2_output(hc2_type *hc2, long tqmin, long tqmax)
{
    int i,  min, pos;
    unsigned int *list;
    //int res = 0;		/* counter for no. of freq. items */
    long cnt;

    //	printf("\nhc2:%d", i);

    list = calloc((int)(1.0/o.thresh), sizeof(unsigned int));
    if (list == NULL) {
        printf("\nCalloc error. ");
        exit(400);
    }

    //printf("\nA_2_List:\n");
    min = (tqmax - tqmin) * o.thresh;
    pos = 0;
    for (i=1; i<M; i++) {
        cnt = hc2_query(hc2, o.tqmin, o.tqmax, i);
        if (cnt > min) {
            pos++;
            list[pos] = i;
            //printf("\nA_2: %ld:%ld", i, cnt);
        }
    }

    list[0] = pos;
    //printf("list = %p\n", list);

    return list;
}

unsigned int *alg3_output(hc_type *titi, long tqmin, long tqmax)
{
    int i,  min, pos;
    unsigned int *list;
    //int res = 0;		/* counter for no. of freq. items */
    long cnt;

    list = calloc((int)(1.0/o.thresh), sizeof(unsigned int));
    if (list == NULL) {
        printf("\nCalloc error. ");
        exit(400);
    }

    //printf("\nA_3_List:\n");
    min = (tqmax - tqmin) * o.thresh;
    pos = 0;
    for (i=1; i<M; i++) {
        cnt = query(titi, o.tqmin, o.tqmax, i);
        if (cnt > min) {

            if(pos+1 > (1.0 / o.thresh) - 1){
                break;
                //assert("err: list overflow for titicount+" && 0);
            }


            pos++;
            list[pos] = i;


            //printf("\nA_3: %ld:%ld", i, cnt);
        }
    }

    list[0] = pos;

    return list;
}



unsigned int *alg8_output(histogram_hc_type *titi, long tqmin, long tqmax)
{
    int i,  min, pos;
    unsigned int *list;
    //int res = 0;		/* counter for no. of freq. items */
    long cnt;

    list = calloc((int)(1.0/o.thresh), sizeof(unsigned int));
    if (list == NULL) {
        printf("\nCalloc error. ");
        exit(400);
    }

    //printf("\nA_3_List:\n");
    min = (tqmax - tqmin) * o.thresh;
    pos = 0;
    for (i=1; i<M; i++) {
        cnt = histogram_query(titi, o.tqmin, o.tqmax, i);
        if (cnt > min) {
            if(pos+1 > (1.0 / o.thresh) - 1){
                break;
                assert("err: list overflow for titicount+h" && 0);
            }

            pos++;
            list[pos] = i;
            //printf("\nA_3: %ld:%ld", i, cnt);
        }
    }

    list[0] = pos;

    return list;
}
unsigned int *alg4_output(linear_type *lc, long tqmin, long tqmax)
{
    int i,  min, pos;
    unsigned int *list;
    //int res = 0;		/* counter for no. of freq. items */
    long cnt;

    list = calloc((int)(1.0/o.thresh), sizeof(unsigned int));
    if (list == NULL) {
        printf("\nCalloc error. ");
        exit(400);
    }

    //printf("\nA_4_List:\n");
    min = (tqmax - tqmin) * o.thresh;
    pos = 0;
    for (i=1; i<M; i++) {
        cnt = linear_query(lc, o.tqmin, o.tqmax, i);
        if (cnt > min) {
            pos++;
            list[pos] = i;
            //printf("\nA_4: %ld:%ld", i, cnt);
        }
    }

    list[0] = pos;

    return list;
}

unsigned int *alg5_output(linearb_type *lc, long tqmin, long tqmax)
{
    int i, min, pos;
    unsigned int *list;
    //int res = 0;		/* counter for no. of freq. items */
    long cnt;

    list = calloc((int)(1.0/o.thresh), sizeof(unsigned int));
    if (list == NULL) {
        printf("\nCalloc error. ");
        exit(400);
    }

    //printf("\nA_5_List:\n");
    min = (tqmax - tqmin) * o.thresh;
    pos = 0;
    for (i=1; i<M; i++) {
        cnt = linearb_query(lc, tqmin, tqmax, i);
        if (cnt > min) {
            pos++;
            list[pos] = i;
            //printf("\nA_5: %ld:%ld", i, cnt);
        }
    }

    list[0] = pos;

    return list;
}


unsigned int *alg6_output(SSW_type *ssw, long tqmin, long tqmax) {
    int i, res;
    unsigned int *list;
    //int res = 0;		/* counter for no. of freq. items */
    SS_bucket *b;
    SS_counter *c;

    SS_type *result = SSW_freqs(ssw,tqmin,tqmax);

    list = calloc((int)(0xffff), sizeof(unsigned int));
    if (list == NULL) {
        printf("\nCalloc error. ");
        exit(400);
    }

    int min = (tqmax - tqmin) * o.thresh;
    res = 0;

    for(b=result->bfirst; b->next != NULL; b = b->next);

    while(b!= NULL) {
        if( b->value > min) {
            for (c=b->ctr; c != NULL; c = c->next) {
                if(res > 1.0 / o.thresh)
                    continue;
                res++;
                list[res] = c->id;
                // printf("at %d : %d ", res, list[res]);
            }
        }
        b = b->prev;
    }

    list[0] = res;	// no. of freq. items found

    //  printf("freq items count = %d\n", res);

    //  printf("OK\n");
    return list;
}


unsigned int *alg7_output(SSW2_type *ssw2, long tqmin, long tqmax) {
    int i, res;
    unsigned int *list;
    SS_bucket *b;
    SS_counter *c;


    SS_type *result = SSW2_freqs(ssw2,tqmin,tqmax); // construct the result inside a SS data structure

    list = calloc((int)(0xffff), sizeof(unsigned int));
    if (list == NULL) {
        printf("\nCalloc error. ");
        exit(400);
    }

    int   min = (tqmax - tqmin) * o.thresh;
    res = 0;

    for(b=result->bfirst; b->next != NULL; b = b->next);

    while(b!= NULL) {
        if( b->value > min) {
            for (c=b->ctr; c != NULL; c = c->next) {
                if(res > 1.0 / o.thresh)
                    continue;
                res++;
                list[res] = c->id;
                // printf("at %d : %d ", res, list[res]);
            }
        }
        b = b->prev;
    }

    list[0] = res;	// no. of freq. items found

    //    printf("freq items count = %d\n", res);

    //  printf("OK\n");
    return list;
}



/******************************************************************************
Check output and calculate precision and recall

******************************************************************************/
void CheckOutput(const char *title, unsigned int *resultlist, int hh, int upt, int outt)
{
    unsigned int i;
    int correct = 0, claimed = 0;
    //unsigned int last = 0;
    double recall, precision;
    int min = (o.tqmax - o.tqmin) * o.thresh;

    if (resultlist[0] > 0) {
        printf("\n%s_Looking at a list of %d items (title, recall, precision, upt, outt)\n", title, resultlist[0]);
        for (i = 1; i <= resultlist[0]; i++) {
            claimed++;
            if (exact[resultlist[i]] > min) {
                correct++;
            }
        }
    } else
        claimed = 1;
    
    if (hh == 0) {
        assert("hh == 0 " && 0);
        printf("%s\t%d\t%d\n", title, upt, outt);
    }
    else {
        recall = 100.0 * ((double) correct) / ((double) hh);
        precision = 100.0 * ((double) correct) / ((double) claimed);
        printf("%s\t%1.2f\t%1.2f\t%d\t%d\n", title, recall, precision, upt, outt);

        printf("#ACCURACY algo= %s precision= %1.2f recall= %1.2f\n", title, precision, recall);
    }
}


/******************************************************************************
Read the stream from file fdata. Reads the first n items.
(For use with real data)

******************************************************************************/
int *ReadStream(char *fname)
{
    char str[512];
    int i;
    int *stream;
    FILE *fin;

    fin=fopen(fname, "r");
    if (fin==NULL) {
        printf("\nData file error.\n");
        exit(20);
    }

    exact = (int *) calloc(M + 1, sizeof(int));
    stream = (int *) calloc(o.range + 1, sizeof(int));
    if (exact==NULL || stream==NULL) {
        printf("calloc error");
        exit(123);
    }

    i=1;
    while (i < o.range) {
        int ret = fscanf(fin, "%s", str);

        if(ret == EOF) {
            o.range = i -1 ;
            break;
        }

        stream[i] = atoi(str);

     //    printf("assigned offset=%d value=%d str=%s\n", i, stream[i], str);

        stream[i] &= M;


        if (stream[i] >= 0) {
            exact[stream[i]]++;
            //  printf("read of %d\n", stream[i]);
        }
        else
            exact[-stream[i]]--;
        i++;
    }

    // printf("\n\n%d -> %s", stream[range-1], str);
    fflush(stdout);
    return(stream);
}

int *ReadStream2(char *fname, int length)
{
    char str[12];
    long i,j;
    int *stream;
    FILE *fin;

    fin=fopen(fname, "r");
    if (fin==NULL) {
        printf("\nData file error.\n");
        exit(20);
    }

    exact = (int *) calloc(M + 1, sizeof(int));
    stream = (int *) calloc(o.range + 1, sizeof(int));
    if (exact==NULL || stream==NULL) {
        printf("calloc error");
        exit(123);
    }



    long a, b;
    long value;
    long lim1, lim2;
    prng_type *prng;
    int netpos;

    prng = prng_Init(time(NULL), 2);
    srand((unsigned) time(NULL));


    netpos = 0;

    beta = 5;// zipfpar;

    int div = length / (beta*o.BSIZE);
    printf("using readstream2 with beta=%d. switching %d times\n", beta, div);

    if(div < 1)
        div = 1;

    if (o.CHANGING_DATA) {
        for (j=0; j<div; j++) {
            a = (long long) (prng_int(prng) % MOD);
            b = (long long) (prng_int(prng) % MOD);

            lim1 = j*length/div + 1;
            lim2 = lim1 + length/div;

            for (i = lim1; i <= lim2; i++) {

                fscanf(fin, "%s", str);
                value = atoi(str);
                value = (hash31(a, b, atoi(str)) & M);

                exact[value]++;
                netpos++;

                if(i > length) {
                    printf("trying to write after end: %d, skipping\n", i); // XXX this is due to a bug of the data stream generator and happens exactly after the last "correct" position. it is a "offset+1 bug" problem
                    continue;
                }

                stream[i] = value;
            }
        }

    } else {

        a = (long long) (prng_int(prng) % MOD);
        b = (long long) (prng_int(prng) % MOD);

        for (i = 1; i <= length; i++) {

            fscanf(fin, "%s", str);
            value = atoi(str);
            value = (hash31(a, b, atoi(str)) & M);

            exact[value]++;
            netpos++;

            stream[i] = value;
        }
    }

    prng_Destroy(prng);
    return (stream);


}


void sort(unsigned int *qfreqs) {
    int i, j, tmp1, tmp2;
    int n = 0xffff;



    for(i = 1; i < n; i++) {
        for(j = i+1; j<= n; j++) {
            if(qfreqs[j*2] > qfreqs[i*2]) {
                tmp1 = qfreqs[i*2-1];
                tmp2 = qfreqs[i*2];
                qfreqs[i*2-1] =qfreqs[j*2-1];
                qfreqs[i*2] =qfreqs[j*2];
                qfreqs[j*2-1] = tmp1;
                qfreqs[j * 2] = tmp2;
            }
        }
    }
}


unsigned int *alg3_get_allfreqs(hc_type *hc, long tqmin, long tqmax) {
    int i;
    unsigned int *qfreqs;

    qfreqs = malloc((0xffff+1) * 2 * sizeof(unsigned int));

    memset(qfreqs, 0, (0xffff+1) * 2 * sizeof(unsigned int));


    for(i = 1; i <= 0xffff; i++) {
        qfreqs[i*2-1] = i;
        qfreqs[i*2] = query(hc, tqmin, tqmax, i);
    }


    return qfreqs;

}



unsigned int *alg8_get_allfreqs(histogram_hc_type *hc, long tqmin, long tqmax) {
    int i;
    unsigned int *qfreqs;

    qfreqs = malloc((0xffff+1) * 2 * sizeof(unsigned int));

    memset(qfreqs, 0, (0xffff+1) * 2 * sizeof(unsigned int));


    for(i = 1; i <= 0xffff; i++) {
        qfreqs[i*2-1] = i;
        qfreqs[i*2] = histogram_query(hc, tqmin, tqmax, i);
    }


    return qfreqs;

}


unsigned int *alg2_get_allfreqs(hc2_type *hc2, long tqmin, long tqmax) {

    int i;
    unsigned int *qfreqs;

    qfreqs = malloc((0xffff+1) * 2 * sizeof(unsigned int));

    memset(qfreqs, 0, (0xffff+1) * 2 * sizeof(unsigned int));


    for(i = 1; i <= 0xffff; i++) {
        qfreqs[i*2-1] = i;
        qfreqs[i*2] = hc2_query(hc2, tqmin, tqmax, i);
    }


    return qfreqs;
}


unsigned int *alg7_get_allfreqs(SSW2_type *ssw2, long tqmin, long tqmax) {
    int i, res;
    unsigned int *qfreqs;
    //int res = 0;		/* counter for no. of freq. items */
    SS_bucket *b;
    SS_counter *c;

    SS_type *result = SSW2_freqs(ssw2,tqmin,tqmax);


    qfreqs = malloc((0xffff+1) * 2 * sizeof(unsigned int));

    memset(qfreqs, 0, (0xffff+1) * 2 * sizeof(unsigned int));

    for(b=result->bfirst; b->next != NULL; b = b->next);


    for(i = 1; i <= 0xffff; i++) {
        qfreqs[i*2-1] = i;
    }

    while(b!= NULL) {
        if( b->value != 0) {
            for (c=b->ctr; c != NULL; c = c->next) {
                //     printf("qfreqs[%d] = %d\n", c->id, b->value);
                qfreqs[c->id*2] = b->value;
            }
        }
        b = b->prev;
    }


    return qfreqs;
}

unsigned int *alg6_get_allfreqs(SSW_type *ssw, long tqmin, long tqmax) {
    int i, res;
    unsigned int *qfreqs;
    //int res = 0;		/* counter for no. of freq. items */
    SS_bucket *b;
    SS_counter *c;

    SS_type *result = SSW_freqs(ssw,tqmin,tqmax);


    qfreqs = malloc((0xffff+1) * 2 * sizeof(unsigned int));

    memset(qfreqs, 0, (0xffff+1) * 2 * sizeof(unsigned int));

    for(b=result->bfirst; b->next != NULL; b = b->next);


    for(i = 1; i <= 0xffff; i++) {
        qfreqs[i*2-1] = i;
    }

    while(b!= NULL) {
        if( b->value != 0) {
            for (c=b->ctr; c != NULL; c = c->next) {
                //     printf("qfreqs[%d] = %d\n", c->id, b->value);
                qfreqs[c->id*2] = b->value;
            }
        }
        b = b->prev;
    }


    return qfreqs;
}


unsigned int *alg0_get_allfreqs() {
    int i, res;
    unsigned int *qfreqs;
    //int res = 0;		/* counter for no. of freq. items */
    SS_bucket *b;
    SS_counter *c;


    qfreqs = malloc((0xffff+1) * 2 * sizeof(unsigned int));

    memset(qfreqs, 0, (0xffff+1) * 2 * sizeof(unsigned int));



    for(i = 1; i <= 0xffff; i++) {
        qfreqs[i*2-1] = i;
    }

    for(i = 1 ; i <= 0xffff; i++) {
        qfreqs[i*2] = exact[i];
    }


    return qfreqs;
}

void print_pr_topk(const char *desc, unsigned int *freqs0, unsigned int *freqs, int k) {

    int i, j;
    double precision;

    precision = 0;


    for(i = 1; i <= k; i++) {

        for(j = 1; j <= k; j++) {
            //   printf("comparing freqs at pos %d,%d: %d,%d\n", i,j,freqs0[2*i-1],freqs[2*j-1]);
            if(freqs0[2*i] < 2) { // 2 because if there are 0,1 occurrences, they are not significative.
                printf("print_pr_topk: there are less frequent items!! k is too high.\n");
                exit(0);
                //XXXXX
            }
            if(freqs0[2*i-1] == freqs[2*j-1]) {
                //    printf("match! %d %d\n", freqs0[2*i-1], freqs[2*j-1]);
                precision++;
            }
        }
    }

    precision /= (double) k ;

    printf("precision of top-%d with %s: %f\n", k, desc, precision);

    //   printf("top-%d freqs:\n", k);
    //   for(i = 1; i <= k; i++) {
    //       printf("\t[i=%d] (id,freq)=(%d,%d)\t\t(trueid,truefreq)=(%d,%d)\n", i, freqs[2*i-1], freqs[2*i],freqs0[2*i-1], freqs0[2*i]);
    //   }


    printf("\n");

}


void print_cumulative_freq_topk(const char *desc, unsigned int *freqs0, unsigned int *freqs, int k, long width) {

    int i, j;
    double precision = 0;
    unsigned long sum = 0;

    for(i = 1; i <= k; i++) {

        if(freqs0[2*i] < 2) { // 2 because if there are 0,1 occurrences, they are not significative.
            printf("print_cumulative_freq_topk: there are less frequent items!! k is too high.\n");
            exit(0);
        }

        for(j = 1; j <= k; j++) {
            if(freqs0[2*i-1] == freqs[2*j-1]) {
                printf("frequent item id=%d freq=%d (exact=%d)\n", freqs[2*j-1], freqs[2*j], freqs0[2*i]);
                precision++;
                sum += freqs[2*i];
            }
        }
    }

    printf("cumulative-frequency of top-%d with %s qwidth=%d batches: %ld\n", k, desc, width, sum);

    printf("#CUMULATIVE_FREQ top= %d algo= %s qwidth= %d nBatches= %ld\n", k, desc, width, sum);


    //   printf("top-%d freqs:\n", k);
    //   for(i = 1; i <= k; i++) {
    //       printf("\t[i=%d] (id,freq)=(%d,%d)\t\t(trueid,truefreq)=(%d,%d)\n", i, freqs[2*i-1], freqs[2*i],freqs0[2*i-1], freqs0[2*i]);
    //   }


}

void experiment_topk( long tqmin, long tqmax) {
    int i;

    N = 0;
    SSW_type *ssw;		//6th Algorithm (SS)
    ssw = SSW_Init(0.00023); // 1950.0/5 XXX it should be /5.0 to use ~ same memory of hc2 (space for buckets not counted, counted just space for counters)
    for (i=1; i<=o.range; i++) {
        SSW_Update(ssw, stream[i], INSERT);
    }


    N = 0;
    SSW2_type *ssw2;		//7th Algorithm (SS2)
    ssw2 = SSW2_Init(0.00023); // XXX it should be /5.0 to use ~ same memory of hc2 (space for buckets not counted, counted just space for counters)
    for (i=1; i<=o.range; i++) {
        SSW2_Update(ssw2, stream[i], INSERT);
    }


    N = 0;
    hc2_type *hc2;		//2nd Algorithm (TiTiCount)
    hc2 = hc2_init(0.99, 0.01, M);
    for (i=1; i<o.range; i++) {
        hc2_update(hc2, stream[i], INSERT);
    }


    N = 0;
    hc_type *titi;		//3rd Algorithm (TiTiCount+)
    titi = hc_init(0.99, 0.01, M);
    for (i=1; i<o.range; i++) {
        hc_update(titi, stream[i], INSERT);
    }


    unsigned int *uilist0 = alg0_get_allfreqs(); // nota: tqmin e tqmax sono considerati perchè exactwin è stata eseguita prima di chiamare questa funzione.
    unsigned int *uilist2 = alg2_get_allfreqs(hc2, tqmin, tqmax);
    unsigned int *uilist3 = alg3_get_allfreqs(titi, tqmin, tqmax);
    unsigned int *uilist6 = alg6_get_allfreqs(ssw,tqmin,tqmax);
    unsigned  int *uilist7 = alg7_get_allfreqs(ssw2, tqmin, tqmax);


    sort(uilist0);
    sort(uilist2);
    sort(uilist3);
    sort(uilist6);
    sort(uilist7);

    int k;


    k = 4;
    print_pr_topk("2_TiTi", uilist0, uilist2, k);
    print_pr_topk("3_TiTi+",uilist0, uilist3, k);
    print_pr_topk("6_SS",uilist0, uilist6, k);
    print_pr_topk("7_SS+",uilist0, uilist7, k);


    k = 16;
    print_pr_topk("2_TiTi", uilist0, uilist2, k);
    print_pr_topk("3_TiTi+",uilist0, uilist3, k);
    print_pr_topk("6_SS",uilist0, uilist6, k);
    print_pr_topk("7_SS+",uilist0, uilist7, k);


    k = 32;
    print_pr_topk("2_TiTi", uilist0, uilist2, k);
    print_pr_topk("3_TiTi+",uilist0, uilist3, k);
    print_pr_topk("6_SS",uilist0, uilist6, k);
    print_pr_topk("7_SS+",uilist0, uilist7, k);


    k = 64;
    print_pr_topk("2_TiTi", uilist0, uilist2, k);
    print_pr_topk("3_TiTi+",uilist0, uilist3, k);
    print_pr_topk("6_SS",uilist0, uilist6, k);
    print_pr_topk("7_SS+",uilist0, uilist7, k);

    /*
    k = 128;
    print_pr_topk("2_TiTi", uilist0, uilist2, k);
    print_pr_topk("3_TiTi+",uilist0, uilist3, k);
    print_pr_topk("6_SS",uilist0, uilist6, k);
    print_pr_topk("7_SS+",uilist0, uilist7, k);
*/
    //   for(i = 1; i <= 0xffff; i++ ) {
    //      printf("id=%d freq=%d\n", uilist2[i*2-1], uilist2[i*2]);
    //  }


}

void experiment_freqs(long tqmin, long tqmax) {
    long i, hh;
    long uptime2, uptime3, uptime4, uptime5,uptime6,uptime7,
    outtime2, outtime3, outtime4, outtime5,outtime6,outtime7;
    unsigned int *uilist;



    hh = ExactWin(tqmin, tqmax);

    printf("hh = %d\n", hh);



    N = 0;
    hc2_type *hc2;		//2nd Algorithm (TiTiCount)
    hc2 = hc2_init(0.99, 0.01, M);
    StartTheClock();
    for (i=1; i<o.range; i++) {
        hc2_update(hc2, stream[i], INSERT);
    }
    uptime2 = StopTheClock();
    StartTheClock();
    uilist = alg2_output(hc2, tqmin, tqmax);
    outtime2 = StopTheClock();
    CheckOutput("2_TiTi", uilist, hh, uptime2, outtime2);
    free(uilist);

    N = 0;
    hc_type *titi;		//3rd Algorithm (TiTiCount+)
    titi = hc_init(0.99, 0.01, M);
    StartTheClock();
    for (i=1; i<o.range; i++) {
        hc_update(titi, stream[i], INSERT);
    }
    uptime3 = StopTheClock();
    StartTheClock();
    uilist = alg3_output(titi, tqmin, tqmax);
    outtime3 = StopTheClock();
    CheckOutput("3_TiTi+", uilist, hh, uptime3, outtime3);
    free(uilist);



    N = 0;
    histogram_hc_type *titih;		//8th Algorithm (TiTiCount+h)
    titih = histogram_hc_init(0.99, 0.0125, M);
    // questo serve per comparare lo stesso run di due tecniche basate su titicount+, dove nella init viene chiamata la random:
    for (i = 0; i < titi->h; i++) {
        titih->a[i] = titi->a[i];
        titih->b[i] = titi->b[i];
    }
    StartTheClock();
    for (i=1; i<o.range; i++) {
        histogram_hc_update(titih, stream[i], INSERT);
    }
    uptime3 = StopTheClock();
    StartTheClock();
    uilist = alg8_output(titih, tqmin, tqmax);
    outtime3 = StopTheClock();
    CheckOutput("8_TiTi+h", uilist, hh, uptime3, outtime3);
    free(uilist);


    N = 0;
    SSW_type *ssw;		//6th Algorithm (SS)
    ssw = SSW_Init(0.0014);
    StartTheClock();
    for (i=1; i<=o.range; i++) {
        SSW_Update(ssw, stream[i], INSERT);
    }
    uptime6 = StopTheClock();
    StartTheClock();
    uilist = alg6_output(ssw, tqmin, tqmax);
    outtime6 = StopTheClock();
    CheckOutput("6_SS", uilist, hh, uptime6, outtime6);
    memset(uilist, 0, sizeof(int) * 0xffff);
    free(uilist);


    N = 0;
    SSW2_type *ssw2;		//7th Algorithm (SS2)
    ssw2 = SSW2_Init(0.0014);
    StartTheClock();
    for (i=1; i<=o.range; i++) {
        SSW2_Update(ssw2, stream[i], INSERT);
    }
    uptime7 = StopTheClock();
    StartTheClock();
    uilist = alg7_output(ssw2, tqmin, tqmax);
    outtime7 = StopTheClock();
    CheckOutput("7_SS+", uilist, hh, uptime7, outtime7);
    memset(uilist, 0, sizeof(int) * 0xffff);
    free(uilist);


    // XXX SS/SS+ have not destroyers

    hc_destroy(titi);
    //     hc2_destroy(hc2);
    //linear_destroy(lc);
    //linearb_destroy(lcb);

}



void experiment_time() {
    long i;
    long updateTime;


    /*
    {
        N = 0;
        hc2_type *hc2;		//2nd Algorithm (TiTiCount)
        hc2 = hc2_init(0.99, 0.01, M);
        StartTheClock();
        for (i=1; i<=range; i++)
            hc2_update(hc2, stream[i], INSERT);
        updateTime = StopTheClock();
        printf("update time of TiTiCount: %ld (%ld)\n", updateTime, range);



        N = 0;
        hc_type *titi;		//3rd Algorithm (TiTiCount+)
        titi = hc_init(0.99, 0.01, M);
        StartTheClock();
        for (i=1; i<=range; i++)
            hc_update(titi, stream[i], INSERT);
        updateTime = StopTheClock();
        printf("update time of TiTiCount+: %ld (%ld)\n", updateTime, range);


        N = 0;
        histogram_hc_type *titih;		//8th Algorithm (TiTiCount+h)
        titih = histogram_hc_init(0.99, 0.0125, M);
        // questo serve per comparare lo stesso run di due tecniche basate su titicount+, dove nella init viene chiamata la random:
        for (i = 0; i < titi->h; i++) {
            titih->a[i] = titi->a[i];
            titih->b[i] = titi->b[i];
        }
        StartTheClock();
        for (i=1; i<=range; i++)
            histogram_hc_update(titih, stream[i], INSERT);
        updateTime = StopTheClock();
        printf("update time of TiTiCount+h: %ld (%ld)\n", updateTime, range);

    }*/


    updateTime = 0;
    N = 0;
    hc2_type *hc2;		//2nd Algorithm (TiTiCount)
    hc2 = hc2_init(0.99, 0.01, M);
    StartTheClock();
    for (i=1; i<=o.range; i++) {
        hc2_update(hc2, stream[i], INSERT);
        if(i % 10000000 == 0) {
            updateTime += StopTheClock(); // milliseconds
            printf("2_tt %ld %ld\n", i, updateTime);
            StartTheClock();
        }
    }


    updateTime = 0;
    N = 0;
    hc_type *titi;		//3rd Algorithm (TiTiCount+)
    titi = hc_init(0.99, 0.01, M);
    StartTheClock();
    for (i=1; i<=o.range; i++) {
        hc_update(titi, stream[i], INSERT);
        if(i % 10000000 == 0) {
            updateTime += StopTheClock();
            printf("3_tt+ %ld %ld\n", i, updateTime);
            StartTheClock();
        }
    }


    updateTime = 0;
    N = 0;
    histogram_hc_type *titih;		//8th Algorithm (TiTiCount+h)
    titih = histogram_hc_init(0.99, 0.0125, M);
    // questo serve per comparare lo stesso run di due tecniche basate su titicount+, dove nella init viene chiamata la random:
    for (i = 0; i < titi->h; i++) {
        titih->a[i] = titi->a[i];
        titih->b[i] = titi->b[i];
    }
    StartTheClock();
    for (i=1; i<=o.range; i++) {
        histogram_hc_update(titih, stream[i], INSERT);
        if(i % 10000000 == 0) {
            updateTime += StopTheClock();
            printf("8_tt+h %ld %ld\n", i, updateTime);
            StartTheClock();
        }
    }


    updateTime = 0;
    N = 0;
    SSW_type *ssw;		//6th Algorithm (SS)
    ssw = SSW_Init(0.0014);
    StartTheClock();
    for (i=1; i<=o.range; i++) {
        SSW_Update(ssw, stream[i], INSERT);
        if(i % 10000000 == 0) {
            updateTime += StopTheClock();
            printf("6_ss %ld %ld\n", i, updateTime);
            StartTheClock();
        }
    }


    updateTime = 0;
    N = 0;
    SSW2_type *ssw2;		//7th Algorithm (SS2)
    ssw2 = SSW2_Init(0.0014);
    StartTheClock();
    for (i=1; i<=o.range; i++) {
        SSW2_Update(ssw2, stream[i], INSERT);
        if(i % 10000000 == 0) {
            updateTime += StopTheClock();
            printf("7_ss+ %ld %ld\n", i, updateTime);
            StartTheClock();
        }
    }


    // XXX SS/SS+ have not destroyers

    // hc_destroy(titi);
    //  hc2_destroy(hc2);
    //linear_destroy(lc);
    //linearb_destroy(lcb);

}



void experiment_topk_freqs() {
    int i;

    /*
    N = 0;
    SSW_type *ssw;		//6th Algorithm (SS)
    ssw = SSW_Init(0.0011); // 1950.0/5 XXX it should be /5.0 to use ~ same memory of hc2 (space for buckets not counted, counted just space for counters)
    for (i=1; i<=range; i++) {
        SSW_Update(ssw, stream[i], INSERT);
    }*/




    N = 0;
    histogram_hc_type *titih;	//8th Algorithm (tt+h)
    titih = histogram_hc_init(0.99, 0.0125, M);
    for (i=1; i<=o.range; i++) {
        histogram_hc_update(titih, stream[i], INSERT);
    }


    N = 0;
    SSW2_type *ssw2;		//7th Algorithm (SS2)
    ssw2 = SSW2_Init(0.0011);
    for (i=1; i<=o.range; i++) {
        SSW2_Update(ssw2, stream[i], INSERT);
    }

    /*
    N = 0;
    hc2_type *hc2;		//2nd Algorithm (TiTiCount)
    hc2 = hc2_init(0.99, 0.01, M);
    for (i=1; i<range; i++) {
        hc2_update(hc2, stream[i], INSERT);
    }
*/

    N = 0;
    hc_type *titi;		//3rd Algorithm (TiTiCount+)
    titi = hc_init(0.99, 0.01, M);
    for (i=1; i<o.range; i++) {
        hc_update(titi, stream[i], INSERT);
    }


    double dummy;
    int query_width;

    for(query_width = 2; query_width <= 16; query_width += 2) {

        //printf("i=%d query_width=%d\n", i, query_width);

        //gen_query(range, &tqmin, &tqmax, &dummy, 1);

        gen_query(o.range, &o.tqmin, &o.tqmax, &dummy, query_width);

//        o.tqmin = 190326;
  //      o.tqmax = 192326; // 16 * 1000;

        printf("query: tqmin=%ld tqmax=%ld\n", o.tqmin, o.tqmax);

        ExactWin(o.tqmin, o.tqmax);

        unsigned int *uilist0 = alg0_get_allfreqs(); // nota: tqmin e tqmax sono considerati perchè exactwin è stata eseguita prima di chiamare questa funzione.
        //unsigned int *uilist2 = alg2_get_allfreqs(hc2, tqmin, tqmax);
        unsigned int *uilist3 = alg3_get_allfreqs(titi, o.tqmin, o.tqmax);
        // unsigned int *uilist6 = alg6_get_allfreqs(ssw,tqmin,tqmax);
        unsigned int *uilist7 = alg7_get_allfreqs(ssw2, o.tqmin, o.tqmax);

        unsigned int *uilist8 = alg8_get_allfreqs(titih, o.tqmin, o.tqmax);


        sort(uilist0);
        //    sort(uilist2);
        sort(uilist3);
        //    sort(uilist6);
        sort(uilist7);

        sort(uilist8);





        int k;

        k = 2;
        print_cumulative_freq_topk("0_exact", uilist0, uilist0, k,query_width);
        //      print_cumulative_freq_topk("2_TiTi", uilist0, uilist2, k,query_width);
        print_cumulative_freq_topk("3_TiTi+",uilist0, uilist3, k,query_width);
        //    print_cumulative_freq_topk("6_SS",uilist0, uilist6, k,query_width);
        print_cumulative_freq_topk("7_SS+",uilist0, uilist7, k,query_width);
        print_cumulative_freq_topk("8_TiTi+h",uilist0, uilist8, k,query_width);


        //   print_pr_topk("6_SS",uilist0, uilist6, k);
        //   print_pr_topk("7_SS+",uilist0, uilist7, k);
        // nota: titicount+ può avere frequenze stimate inferiori alla stima reale per via di distribuzioni non uniformi nei dati. solo in hCount "puro" cè solo sovrastima, in titicount+ ci può essere anche sottostima.
        //        int cnt = query(titi, tqmin, tqmax, 3);
        //      printf("cnt = %d\n", cnt);
        //    printf("exact = %d\n", exact[3]);

        k = 3;
        print_cumulative_freq_topk("0_exact", uilist0, uilist0, k,query_width);
        //    print_cumulative_freq_topk("2_TiTi", uilist0, uilist2, k,query_width);
        print_cumulative_freq_topk("3_TiTi+",uilist0, uilist3, k,query_width);
        //  print_cumulative_freq_topk("6_SS",uilist0, uilist6, k,query_width);
        print_cumulative_freq_topk("7_SS+",uilist0, uilist7, k,query_width);
        print_cumulative_freq_topk("8_TiTi+h",uilist0, uilist8, k,query_width);

        //   print_pr_topk("6_SS",uilist0, uilist6, k);
        //   print_pr_topk("7_SS+",uilist0, uilist7, k);


        k = 4;
        print_cumulative_freq_topk("0_exact", uilist0, uilist0, k,query_width);
        //  print_cumulative_freq_topk("2_TiTi", uilist0, uilist2, k,query_width);
        print_cumulative_freq_topk("3_TiTi+",uilist0, uilist3, k,query_width);
        //    print_cumulative_freq_topk("6_SS",uilist0, uilist6, k,query_width);
        print_cumulative_freq_topk("7_SS+",uilist0, uilist7, k,query_width);
        print_cumulative_freq_topk("8_TiTi+h",uilist0, uilist8, k,query_width);


        k = 8;
        print_cumulative_freq_topk("0_exact", uilist0, uilist0, k,query_width);
        //  print_cumulative_freq_topk("2_TiTi", uilist0, uilist2, k,query_width);
        print_cumulative_freq_topk("3_TiTi+",uilist0, uilist3, k,query_width);
        //    print_cumulative_freq_topk("6_SS",uilist0, uilist6, k,query_width);
        print_cumulative_freq_topk("7_SS+",uilist0, uilist7, k,query_width);
        print_cumulative_freq_topk("8_TiTi+h",uilist0, uilist8, k,query_width);



        free(uilist8);

        free(uilist0);
        //    free(uilist2);
        free(uilist3);
        //    free(uilist6);
        free(uilist7);
    }
}


void experiment_zipf() {
    unsigned int *uilist0 = alg0_get_allfreqs(); // nota: tqmin e tqmax sono considerati perchè exactwin è stata eseguita prima di chiamare questa funzione.
    sort(uilist0);

    int i;

    for(i = 1; i <= 0xffff; i++) {
        printf("frequency: %d %d\n",i-1, uilist0[i*2]);
    }

}




//===========================================================================
//=  Function to generate Zipf (power law) distributed random variables     =
//=    - Input: alpha and N                                                 =
//=    - Output: Returns with Zipf distributed random variable              =
//===========================================================================
int zipf(double alpha, int n)
{
    static int first = TRUE;      // Static first time flag
    static double c = 0;          // Normalization constant
    double z;                     // Uniform random number (0 < z < 1)
    double sum_prob;              // Sum of probabilities
    double zipf_value;            // Computed exponential value to be returned
    int    i;                     // Loop counter
    // Compute normalization constant on first call only - or with magic params

    if(alpha == 0 && n == 0) {
        c = 0;
        first = TRUE;
        return -1;
    }

    if (first == TRUE)
    {
        for (i=1; i<=n; i++)
            c = c + (1.0 / pow((double) i, alpha));
        c = 1.0 / c;
        first = FALSE;
    }


    // Pull a uniform random number (0 < z < 1)
    do
    {
        z = rand_val(0);
    }
    while ((z == 0) || (z == 1));
    // Map z to the value
    sum_prob = 0;
    for (i=1; i<=n; i++)
    {
        sum_prob = sum_prob + c / pow((double) i, alpha);
        if (sum_prob >= z)
        {
            zipf_value = i;
            break;
        }
    }
    // Assert that zipf_value is between 1 and N
    assert((zipf_value >=1) && (zipf_value <= n));
    return(zipf_value);
}
//=========================================================================
//= Multiplicative LCG for generating uniform(0.0, 1.0) random numbers    =
//=   - x_n = 7^5*x_(n-1)mod(2^31 - 1)                                    =
//=   - With x seeded to 1 the 10000th x value should be 1043618065       =
//=   - From R. Jain, "The Art of Computer Systems Performance Analysis," =
//=     John Wiley & Sons, 1991. (Page 443, Figure 26.2)                  =
//=========================================================================
double rand_val(int seed)
{
    const long  a =      16807;  // Multiplier
    const long  m = 2147483647;  // Modulus
    const long  q =     127773;  // m div a
    const long  r =       2836;  // m mod a
    static long x;               // Random int value
    long        x_div_q;         // x divided by q
    long        x_mod_q;         // x modulo q
    long        x_new;           // New x value
    // Set the seed if argument is non-zero and then return zero
    if (seed > 0)
    {
        x = seed;
        return(0.0);
    }
    // RNG using integer arithmetic
    x_div_q = x / q;
    x_mod_q = x % q;
    x_new = (a * x_mod_q) - (r * x_div_q);
    if (x_new > 0)
        x = x_new;
    else
        x = x_new + m;
    // Return a random value between 0.0 and 1.0
    return((double) x / m);
}
