#include "ssw.h"

#include <stdio.h>
#include <stdlib.h>
//#include <malloc.h>
#include <math.h>
#include <string.h>

extern long N;		// current transaction number
extern long BSIZE;	// batch size


SSW_type  *SSW_Init(double epsilon)
{
    int i;

    SSW_type *result = (SSW_type *)calloc(1, sizeof(SSW_type));

    result->epsilon = epsilon;

    for(i = 0; i < HWINDOWS; i++) {
        result->ss[i] = SS_Init(epsilon);
    }

    //  printf("SS using %d counters.\n", result->ss[0]->numc);

    return result;
}


void shift(SSW_type *ssw) {
    int i,j;

    int bn = N/BSIZE;

    for(i = HWINDOWS-2;i>0;i--) {

        if(bn % ((int)pow(2,i)) == 0) {

            SS_Destroy(ssw->ss[i+1]);
            ssw->ss[i+1] = SS_Init(ssw->epsilon);

            SS_bucket *b;
            SS_counter *c;

            //   int cnt = 0;

            for(j  = 0; j <= i; j++) {
                for (b=ssw->ss[j]->bfirst; b != NULL; b = b->next) {
                    if(b->value != 0)
                        for (c=b->ctr; c != NULL; c = c->next) {

                        // cnt += b->value;
                        SS_Add(ssw->ss[i+1], c->id, c->bucket->value);
                    }
                }
            }

            //      printf("inc to bucket %d cnt: %d\n", i+1, cnt);
            //      printf("bucket %d is empty: %d\n", i+i, ssw->ss[i+1]->empty);

            break;
        }
    }
}


void SSW_Update(SSW_type *ssw, int item, int ttype)
{
    int i,j, reminder;

    N++;

    if(N % BSIZE == 0) {
        shift(ssw);
        reminder = N % (2 * BSIZE);
        if(reminder < BSIZE) {
            SS_Destroy(ssw->ss[0]);
            ssw->ss[0] = SS_Init(ssw->epsilon);
        } else {
            SS_Destroy(ssw->ss[1]);
            ssw->ss[1] = SS_Init(ssw->epsilon);
        }
    }

    if (N % (2 * BSIZE) < BSIZE) {
        SS_Update(ssw->ss[0],item,ttype);
    }
    else {
        SS_Update(ssw->ss[1],item,ttype);
    }

}


/*
 // my original function.
void SSW_Update(SSW_type *ssw, int item, int ttype)
{
    int i,j;

    N++;
    // printf("at %d val = %d\n", N, item);

    if (N % (2 * BSIZE) < BSIZE) {
        SS_Update(ssw->ss[0],item,ttype);
    }
    else {
        SS_Update(ssw->ss[1],item,ttype);
    }

    int bn = N/BSIZE;

    if (N % (BSIZE*2) == 0) { // shift!


        for(i = HWINDOWS-2;i>=0;i--) {

            if(bn % ((int)pow(2,i)) == 0) {

                //     printf("N = %d BN = %d shifting 0...%d to %d\n", N, bn, i, i+1);

                SS_Destroy(ssw->ss[i+1]);
                ssw->ss[i+1] = SS_Init(ssw->epsilon);

                SS_bucket *b;
                SS_counter *c;

                //   int cnt = 0;

                for(j  = 0; j <= i; j++) {
                    for (b=ssw->ss[j]->bfirst; b != NULL; b = b->next) {
                        if(b->value != 0)
                            for (c=b->ctr; c != NULL; c = c->next) {

                            // cnt += b->value;
                            SS_Add(ssw->ss[i+1], c->id, c->bucket->value);
                        }
                    }
                }

                //      printf("inc to bucket %d cnt: %d\n", i+1, cnt);
                //      printf("bucket %d is empty: %d\n", i+i, ssw->ss[i+1]->empty);

                break;
            }
        }

        if(bn % 2 == 0) { // if bn even
            SS_Destroy(ssw->ss[0]);
            ssw->ss[0] = SS_Init(ssw->epsilon);
        } else {
            SS_Destroy(ssw->ss[1]);
            ssw->ss[1] = SS_Init(ssw->epsilon);
        }

    }

}

*/


/*
 // my original function
SS_type *SSW_freqs(SSW_type *ssw, long tqmin, long tqmax)
{
    SS_bucket *b;
    SS_counter *c;

    int i;

    //   for(i = HWINDOWS-1; i>=0;i--) {
    //       printf("bucket %d empty: %d\n", i,ssw->ss[i]->empty);
    //   }


    SS_type *result = SS_Init(1.0/(double)0xffff); // result is exact. no approximation introduced here. note that with hcount there isn't this problem.

    //  printf("counters of result: %d\n", result->numc);

    int bn = N / BSIZE; // n of buckets
    int bw; // bucket width
    double wf; // weighted fraction
    int seen = 0; // n of transactions seen so far

    int qw = 0;
    int started = 0;

    int ready =0;

    for(i = HWINDOWS-2; i>=0;i--) {

        // now we consider bucket i+1

        if(!ready && (bn % ((int)pow(2,i+1))) > pow(2,i)) // is bucket i+1 used?
            ready = 1;


        //   printf("constructing result at bucket %d: ready=%d started=%d qw=%d seen=%d\n", i+1, ready, started,qw, seen);

        if(!ready)
            continue;

        // here, we know that bucket i+1 has been used.

        // printf("bn = %d i=%d\n", bn, i);

        bw = pow(2,i)*BSIZE; // larghezza bucket i+1-esimo (non i-esimo!)
        seen += bw;

        //   printf( "bucket %d bw = %d => seen = %d\n" ,i+1, bw,seen);

        if(!started && seen >= tqmin ) {

            //  printf("CASE 1\n");

            wf = ((double)(seen-tqmin))/(double)bw;

            //printf("checkkkkkk seen=%d tqmax=%d\n", seen, tqmax);
            if(tqmax <= seen) {
                //    printf("CASE 1.3\n");
                wf -= ((double)(seen-tqmax))/(double)bw;
            }

            qw += bw * wf;

            //     printf("wf=%f => qw = %d\n",wf,qw);

            if(wf < 0 || wf > 1) {
                printf("err: wf < 0 > 1 (1)\n");
                exit(1);
            }

            if(ssw->ss[i+1]->empty) {
                printf("looking inside empty bucket\n");
                exit(1);
            }

            for (b=ssw->ss[i+1]->bfirst; b != NULL; b = b->next) {
                if(b->value != 0)
                    for (c=b->ctr; c != NULL; c = c->next) {
                    SS_Add(result, c->id, c->bucket->value * wf);
                }
            }
            started = 1;

            if(seen >= tqmax) {
                break;
            }

            continue;
        }


        if(started && seen <= tqmax) {

            //   printf("CASE 2\n");

            if(ssw->ss[i+1]->empty) {
                printf("looking inside empty bucket\n");
                exit(1);
            }
            qw += bw;
            for (b=ssw->ss[i+1]->bfirst; b != NULL; b = b->next) {
                if(b->value != 0)
                    for (c=b->ctr; c != NULL; c = c->next) {
                    SS_Add(result, c->id, c->bucket->value);
                }
            }
            continue;
        }


        if(started) { // we are ending


            //       printf("CASE 3\n");

            wf = 1.0-((double)(seen-tqmax))/(double)bw;

            if(wf < 0 || wf > 1) {
                printf("err: wf < 0 > 1 (2)\n");
                printf("tqmax=%d seen=%d\n", tqmax, seen);
                exit(1);
            }

            qw += bw * wf;


            //      printf("wf=%f => qw = %d\n",wf,qw);
            if(ssw->ss[i+1]->empty) {
                printf("looking inside empty bucket\n");
                exit(1);
            }
            for (b=ssw->ss[i+1]->bfirst; b != NULL; b = b->next) {
                if(b->value != 0)
                    for (c=b->ctr; c != NULL; c = c->next) {
                    SS_Add(result, c->id, c->bucket->value * wf);
                }
            }

            break;
        }
    }

    //  printf("ok, returning result.\n");
    return result;

}



*/


typedef struct CounterInfo {
    long startIndex;
    long endIndex;
    int started;
    int finished;
    long currentSum;
} CounterInfo;
#define TRUE	1
#define FALSE	0

void SSW_update_freqs(SSW_type *ssw, SS_type *result, double frac, int wid) {
    SS_bucket *b;
    SS_counter *c;

    for (b=ssw->ss[wid]->bfirst; b != NULL; b = b->next) {
        if(b->value != 0)
            for (c=b->ctr; c != NULL; c = c->next) {
            SS_Add(result, c->id, c->bucket->value * frac);
        }
    }


}

void SSW_calculateSum(int counterIndex, long counterSize, SSW_type *result, SSW_type *ssw, CounterInfo *info)
{
    double startFraction, finalFraction;

    if (info->finished)
        return;

    info->currentSum += counterSize;
    //printf("info: %ld %ld %ld %d\n", info->currentSum, info->startIndex, counterSize, info->started);
    //starting the count, getting start fraction and possibly end fraction
    if (!info->started && (info->currentSum > info->startIndex)) {
        info->started = TRUE;

        startFraction =	(double) (info->currentSum - info->startIndex) / (double) counterSize;
        //printf("data: %ld %ld %ld %f\n", info->currentSum, info->startIndex, counterSize, startFraction);
        if (info->started && (info->currentSum > info->endIndex)) {
            info->finished = TRUE;

            if (info->endIndex >= N)
                return SSW_update_freqs(ssw, result, 1.0, counterIndex);
                // return c[counterIndex];

            finalFraction = (double) (info->endIndex - (info->currentSum -  counterSize)) / (double) counterSize;
            return SSW_update_freqs(ssw, result, (startFraction - (1 - finalFraction)), counterIndex);
            //return (long)((startFraction - (1 - finalFraction)) *  c[counterIndex]);
        }
         return SSW_update_freqs(ssw, result, startFraction, counterIndex);
        //return (long)(startFraction * c[counterIndex]);
    }
    //getting end fraction for blocks starting from beginning of block
    if (info->started && (info->currentSum > info->endIndex)) {
        info->finished = TRUE;

        if (info->endIndex >= N)
            return SSW_update_freqs(ssw, result, 1.0, counterIndex);
        // return c[counterIndex];

        finalFraction = (double) (info->endIndex - (info->currentSum - counterSize)) / (double) counterSize;
        return SSW_update_freqs(ssw, result, finalFraction, counterIndex);
        // return (long)(finalFraction * c[counterIndex]);
    }
    //getting full blocks for blocks in the middle
    if (info->started && (info->currentSum <= info->endIndex)) {
        return SSW_update_freqs(ssw, result, finalFraction, counterIndex);
        //return c[counterIndex];
    }
    return;
}

SS_type *SSW_freqs(SSW_type *ssw, long tqmin, long tqmax)
{
    long i,j;
    int powerValue;
    long finalCount, counterSize;
    CounterInfo info;
    SS_type *result = SS_Init(1.0/(double)0xffff);

    info.startIndex = tqmin;
    info.endIndex = tqmax;
    info.currentSum = 0;
    info.started = 0;
    info.finished = 0;

    finalCount = 0;
    counterSize = 0;
    powerValue = 0;
    //printf("data: %ld %ld %ld\n", c->numOfCounters, c->current, c->blockSize);
    for (powerValue = HWINDOWS-2; powerValue > 0; powerValue--) {
        counterSize = (int) pow(2.0, (double) powerValue) * BSIZE;
        //System.out.println("threshold: " + counterSize + "current:" + counter + "mod " + counter%(2*counterSize));
        if (N % (2 * counterSize) >= counterSize) {
            SSW_calculateSum(powerValue + 1, counterSize, result, ssw, &info);
        }
    }
    //printf("finalCount1: %ld\n", finalCount);
    //for 1a
    counterSize = BSIZE;
    SSW_calculateSum(0, counterSize, result, ssw, &info);

    // printf("finalCount2: %ld\n", finalCount);
    //for 1b
    if (N % (2 * counterSize) >= counterSize) {
        SSW_calculateSum(1, counterSize, result, ssw, &info);
        //printf("finalCount3: %ld\n", finalCount);
    }

    return result;
}


